/* trans.c - a helper program for IPOFIX.  By Jeff Winston  1/14/00 Rev. 1.01
Copyright 1999, 2000

---------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.  or check out http://www.gnu.org/copyleft/gpl.html
---------------------------------------------------------------------
*/

# include  <stdio.h>
# include  <string.h>

/* The user must edit 3 lines below for the project (search for "project-specific" */

# define MAX_HIERARCHY      20    /* max levels of hierarchy in an hierarchical instance name */
# define MAXTOKENS       10000    /* max number of tokens in a verilog statement */
# define MAX_MODULES      2000    /* maximum number of modules in the input file */
# define MAX_MOD_CALLS    5000    /* maximum number of module calls to other non-primitive modules in the input file */
# define MAX_GATES       200000   /* maximum number of gates in the design */

char 
     instr[255],	/* input line */
     fullstr[655350],	/* complete parsed verilog line */ 
     tmpstr[255],	/* general temp string */
     modtmp[64],	/* strings to hold module name */
     modname[32], 

     modnames[MAX_MODULES][32];	 /* list of all modules */
long modcnt;			 /* tracks size of array used */

char i_mod[MAX_MOD_CALLS][32],	 /* module reference database */
     i_cmod[MAX_MOD_CALLS][32],
     i_inst[MAX_MOD_CALLS][32];
long i_cnt;			/* tracks size of array used */


char tokens[MAXTOKENS][64],	 /* tokens in a verilog statement */
     itokens[MAX_HIERARCHY][64]; /* used to decompose hierarchical names */
long ntk, intk;			 /* tracks # of tokens filled */

FILE *infile_ptr,  *transfile_ptr, *outfile_ptr;

char  infile[80],   transfile[80],  outfile[80];

char gates[MAX_GATES][20][32];	    /* Gates database */
long modnum[MAX_GATES];		    /* number of module containing gate */
long tokens_used[MAX_GATES];	    /* number of tokens in gates[x] used */
long gatecnt;			    /* tracks size of array used */

long diag,	        /* diag flag */
     x,			/* generic loop variable */
     done,		/* generic done flag */
     all_done,		/* used for string assembly */
     parsing_comment,
     modnumber,		/* module number of current gate in module database */
     gatenumber,	/* number of current gate in gate database*/
     ilevel,		/* index through itokens */
     finished,		/* flag to track when we're done with transition section of file */
     found_cnt,		/* counts how many names we could identify */
     punt_cnt;		/* counts how many we had to punt to primetime */


main(argc,argv)
int argc;                             /* Parse command line */
char *argv[];
{
    diag=0;
  
    if ((argc !=4) && (argc !=5))	
    {
        printf("Error:  Incorrect number of parameters on command line\n");
	printf("Enter input filename, transition filename, output filename [diag level]\n");
        exit(0);
    }

    ++argv;
    strcpy(infile,*argv);
    ++argv;
    strcpy(transfile,*argv);
    ++argv;
    strcpy(outfile,*argv);
    if (argc==5)
    {
	++argv;
	strcpy(tmpstr,*argv);
	sscanf(tmpstr,"%d",&diag);
    }
					/* try to open files */

    if ( (infile_ptr=fopen(infile, "r")) ==NULL )
    {
        printf("Error:  Error opening input file /%s/\n\n",infile);
        exit(1);
    }
    if ( (transfile_ptr=fopen(transfile, "r")) ==NULL )
    {
        printf("Error:  Error opening input file /%s/\n\n",infile);
        exit(1);
    }
    if ((outfile_ptr=fopen(outfile, "w")) ==NULL)
    {
        printf("Error:  Error opening output file /%s/\n\n",outfile);
        exit(1);
    }

								/* PROJECT-SPECIFIC commands for P T */
    fprintf(outfile_ptr,"read_verilog ./sdf/phoenix_foripo.v\n");  /* your netlist */
    fprintf(outfile_ptr,"link_design phoenix\n");                  /* your top level module */
    fprintf(outfile_ptr,"# $ Transitions phoenix 3.0 3.0\n");      /* your desired command line to IPOFIX */

    parsing_comment=0;
    all_done=1;
    done=0;
    gatecnt=0;
    modcnt=0;			/* build databases from input file */
    i_cnt=0;
    while (fgets(instr,255,infile_ptr) != NULL)             /* Keep getting lines until no more data */
    {
       ntk = get_tokens(instr);
       if (ntk > 1) 
       {						/* build full lines out of incomplete ones */
	 done = add_str(instr);
	 while (done == 0) 
	 {
	      if (fgets(instr,255,infile_ptr) == NULL) done=1;
	      else done = add_str(instr);
	 }
	 done=0;
	 ntk = get_tokens(fullstr);			/* parse the full line */


	 if (strcmp(tokens[0],"module")==0)		/* module list */
	 {
		strcpy(modnames[modcnt],tokens[1]);
		printf("Reading module %s\n",tokens[1]);fflush(stdout);
		modcnt++;	     
	        limit_check(modcnt,MAX_MODULES,"Maximum Number of Modules");

         }
	 else if (strcmp(tokens[2],"(")==0)   /* if it's a call */
	 {
	    for (x=0; x<(modcnt-1);x++)		/* if its a call to a known module - add to module call database */
	    {
		if (strcmp(modnames[x],tokens[0])==0)
		{
			strcpy(i_mod[i_cnt],  modnames[modcnt-1]);
			strcpy(i_cmod[i_cnt], tokens[0]);
			strcpy(i_inst[i_cnt], tokens[1]);
			i_cnt++;
			limit_check(i_cnt,MAX_MOD_CALLS,"Maximum Module References");
			x = modcnt;
		}
	   }
	   ntk = get_tokens_no_paren(fullstr);	/* re-parse it with no punctuation */
						/* add to gate database */

	   modnum[gatecnt]=modcnt-1;
	   tokens_used[gatecnt] = ntk;
	   if (tokens_used[gatecnt] > 20) tokens_used[gatecnt] = 20;
	   for (x=0;x<tokens_used[gatecnt];x++)
	   {
		if (strlen(tokens[x])>32)
		{
			printf("Gate Token %s too long (32 char max), exiting\n",tokens[x]);
			exit(0);
		}
		strcpy(gates[gatecnt][x],tokens[x]);
	   }
	   gatecnt++;
	   limit_check(gatecnt,MAX_GATES,"Maximum Number of Gates");
        }
     }
   }	
   printf("Loaded %d modules, %d module references, %d gates. Processing transitions file...\n",modcnt,i_cnt,gatecnt);fflush(stdout);

   /* process transition file */


   found_cnt=0;punt_cnt=0; finished=0;
   while ((finished ==0) && (fgets(instr,255,transfile_ptr) != NULL))             /* Keep getting lines until no more data */
   {
       ntk = get_tokens(instr);

       /* look for first transition */
       if ((ntk > 0) && (strcmp(tokens[0],"Pin:")==0) )
       {
	        if (diag>0) {printf("Processing line %s...",instr);fflush(stdout);}
		intk = get_tokens_hier(tokens[1]);

		/* walk hierarchy to get source module */
		strcpy(modtmp,modnames[modcnt-1]);
		ilevel=0;
		while (intk >2)
		{
			done=0;
			for (x=0; x<i_cnt;x++)
			{
				if ((strcmp(i_mod[x],modtmp)==0) && (strcmp(i_inst[x],itokens[ilevel])==0)) 
				{
					if (diag>0) 
					   printf("For module %s, found that instance name %s points to module %s\n",
						    modtmp,itokens[ilevel],i_cmod[x]);

					strcpy(modtmp,i_cmod[x]);
					x = i_cnt+10;
					ilevel++;
					done=1;
					intk--;
				}
			}		
			if (done==0)
			{
			        printf("Error: For module %s, couldn't find module name for instance %s in line %s\n",
						    modtmp,itokens[ilevel], instr);
				exit(1);
			}
			
		}
		done=0;
		if (diag>0) {printf("module: %s...",modtmp);fflush(stdout);}

		/* get module number of module */
		for (x=0;x<modcnt;x++)
		{
			if (strcmp(modnames[x],modtmp)==0)
			{
				done=1;
				modnumber=x;
				x=modcnt;
			}
		}
		if (done==0)
		{
			printf("Error: couldn't find module %s in database while processing line %s",modtmp,instr);
			exit(0);
		}
		done=0;

		/* search gate database for instance name */
		for (x=0;x<gatecnt;x++)
		{
			if (modnumber == modnum[x])
			{
				if (strcmp(gates[x][1],itokens[ilevel])==0)
				{
					done=1;
					gatenumber=x;
					x=gatecnt;
				}
			}
		}
		if (done==0)
		{
			printf("Couldn't find instance %s in module %s\n",itokens[ilevel],modnames[modnumber]);			
		}
		
		strcpy(tmpstr,tokens[1]);	/* store reference name */

		/* skip down 3 lines and get tokens */
		for (x=0; x<3; x++)
		{
			if (finished==0) finished = (fgets(instr,255,transfile_ptr) == NULL);
		}
		ntk = get_tokens(instr);
		
		/* if it's not a transition-related line, we're done */

		if ((finished==1) || (strcmp(tokens[1],"Transition")!=0))
		{
			finished=1;
			if (strcmp(tokens[1],"Fanout")==0)
				     printf("Reached first Fanout line, all done!\n");
				else printf("Error:  bad line: %s",instr);
		}
		else
		{				/* if it's an indentifiable output, we found it */
			if ((done==1) && (
			((strcmp(itokens[ilevel+1],"z")==0) ||
			(strcmp(itokens[ilevel+1],"zn")==0) ||
			(strcmp(itokens[ilevel+1],"q")==0) ||
			(strcmp(itokens[ilevel+1],"qn")==0) ||
			(strcmp(itokens[ilevel+1],"cin")==0) ||
			(strcmp(itokens[ilevel+1],"so")==0) ||
			(strcmp(itokens[ilevel+1],"co")==0) ||
			( (strcmp(itokens[ilevel+1],"s")==0) && ( (find_string(gates[gatenumber][0],"mx21d")==0) 
		  					&&   (find_string(gates[gatenumber][0],"mi21d")==0)) ) )))
			{
				found_cnt++;
				fprintf(outfile_ptr,"# TRANSDATA %s %s %s\n",tmpstr,gates[gatenumber][0],tokens[3]);
			}
			else	/* let primetime find it */
			{
				punt_cnt++;
				fprintf(outfile_ptr,"# FORTRANSFIX %s\n",tokens[3]);
				fprintf(outfile_ptr,"report_transitive_fanin -to %s\n",tmpstr);
			}
		}
	   }
   }
   fprintf(outfile_ptr,"# $ END_CHANGES_FILE\n");
   fprintf(outfile_ptr,"quit\n");
   printf("%d supplied, %d punted to Primetime\n",found_cnt,punt_cnt);
   printf("%s written\n",outfile);
}


int add_str(str)
char *str;
{
	char tmpstr[65536];
	long loc, len, sp,w;
	len = strlen(str);
	if (all_done == 1)
        {
	   strcpy(fullstr,"\0");
  	   all_done=0;
	}
	if (len>0) len--;
	for (loc=0; loc < len; loc++)
	{
		if ( (parsing_comment==1) && (str[loc]=='*') && (str[loc+1]=='/') )
		{
		  parsing_comment=0;
		  loc++;
		}		  
		else if (parsing_comment==0)
		{
		         if ((str[loc]=='/') && (str[loc+1]=='*'))  parsing_comment=1;
		    else if ((str[loc]=='/') && (str[loc+1]=='/'))  loc = len;
		    else if (str[loc]=='(')   strncat(fullstr," ( ",3); 
		    else if (str[loc]==')')   strncat(fullstr," ) ",3); 
		    else if (str[loc]=='{')   strncat(fullstr," { ",3); 
		    else if (str[loc]=='}')   strncat(fullstr," } ",3); 
		    else if (str[loc]==';') 
		    {
		        strncat(fullstr,&str[loc],1); 
		        strncat(fullstr,"\0",1);
		        all_done=1;
                    }			       
		    else if  (
			 (loc+8 < len)     &&
			 (str[loc+0]=='e') &&
			 (str[loc+1]=='n') &&
			 (str[loc+2]=='d') &&
			 (str[loc+3]=='m') &&
			 (str[loc+4]=='o') &&
			 (str[loc+5]=='d') &&
			 (str[loc+6]=='u') &&
			 (str[loc+7]=='l') &&
			 (str[loc+8]=='e') )
		    {
		        for (w=loc;w<loc+9;w++)
		        {
		            strncat(fullstr,&str[w],1); 
		        }
		        strncat(fullstr,"\0",1);
		        all_done=1;
                    }			       
		    else  strncat(fullstr,&str[loc],1); 
	       }	        
	}	        
	if (all_done==1)
	{
	    sp=0;
	    strcpy(tmpstr,fullstr);
	    strcpy(fullstr,"\0");
	    len = strlen(tmpstr);
	    for (w=0;w<len;w++)
	    {
		if ((tmpstr[w] != ' ') || (sp==0))
		{
		    strncat(fullstr,&tmpstr[w],1); 
		    if (tmpstr[w] == ' ') sp=1; else sp=0;
		}
	    }
 	    strncat(fullstr,"\0",1);
	}
	return(all_done);
}



int get_tokens_no_paren(str)           /* get tokens on line. Put in global tokens[32][32], return  count */
char *str;
{
   int len, loc, cnt, sp, start;

   len = strlen(str);  sp=1; cnt=0; 

   for (loc=0; loc <= len; loc++)
   {
     
       if ( (loc < len) && (str[loc] != ' ') && (str[loc] != '	') && (str[loc] != '.') && (str[loc] != '(') && (str[loc] != '\n') 
			&& (str[loc] != ')') && (str[loc] != ',') && (str[loc] != ';') && (str[loc] != "\0") )
       {
           if (sp==1)		/* if state = waiting */
           {
               cnt++;		/* incr token count, flip state to running, assign start */
               sp=0;
               start=loc;
           }
       }
       else		/* if not valid character */
       {
           if (sp==0)	/* if state = running, flip state to waiting, append good characters to string */
           {
               sp=1;
               strncpy(tokens[cnt-1],&str[start],loc-start);
               strncpy(&tokens[cnt-1][loc-start],"\0",1);

           }
       }
   }
   return(cnt);
}
int get_tokens(str)           /* get tokens on line. Put in global tokens[32][32], return  count */
char *str;
{
   int len, loc, cnt, sp, start;

   len = strlen(str);  sp=1; cnt=0; 

   for (loc=0; loc <= len; loc++)
   {
     
       if ( (loc < len) && (str[loc] != ' ') && (str[loc] != '	') && (str[loc] != '\n') && (str[loc] != "\0") )
       {
           if (sp==1)		/* if state = waiting */
           {
               cnt++;		/* incr token count, flip state to running, assign start */
               sp=0;
               start=loc;
           }
       }
       else		/* if not valid character */
       {
           if (sp==0)	/* if state = running, flip state to waiting, append good characters to string */
           {
               sp=1;
               strncpy(tokens[cnt-1],&str[start],loc-start);
               strncpy(&tokens[cnt-1][loc-start],"\0",1);

           }
       }
   }
   return(cnt);
}
long find_string(field, source)         /* returns LOC */
char *field;
char *source;
{
    long flen;
    long slen; 
    char    *sPtr;
    long loc;

    flen = strlen(field);
    slen = strlen(source);
    if (slen > flen) return (0);     /* error */
    else
    {
        for (loc=0, sPtr = field; loc <= flen-slen; loc++, sPtr++)
        {
            if (strncmp(source,sPtr, slen) ==0) return(loc+1);
        }
    }
    return (0);
}
int get_tokens_hier(str)           /* get tokens on line. Put in global tokens[32][32], return  count */
char *str;
{
   int len, loc, cnt, sp, start;

   len = strlen(str);  sp=1; cnt=0; 

   for (loc=0; loc <= len; loc++)
   {
     
       if ( (loc < len) && (str[loc] != ' ') && (str[loc] != '\n') && (str[loc] != '/') && (str[loc] != "\0") )
       {
           if (sp==1)		/* if state = waiting */
           {
               cnt++;		/* incr token count, flip state to running, assign start */
	       limit_check(cnt,MAX_HIERARCHY,"Maximum number of tokens on line");
               sp=0;
               start=loc;
           }
       }
       else		/* if not valid character */
       {
           if (sp==0)	/* if state = running, flip state to waiting, append good characters to string */
           {
               sp=1;
               strncpy(itokens[cnt-1],&str[start],loc-start);
               strncpy(&itokens[cnt-1][loc-start],"\0",1);

           }
       }
   }
   return(cnt);
}
limit_check(value,limit,errstr)
long value, limit; char *errstr; 
{
	if (value >limit -1)
	{
		printf("Error: Exceeded %s (%d).  Exiting...\n",errstr,limit); 
		exit(1);
	}
}
